/*
 * Copyright (c) 2008, 2010, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores
 * CA 94065 USA or visit www.oracle.com if you need additional information or
 * have any questions.
 */
package com.sun.lwuit.impl.android;

import android.view.KeyEvent;
import com.sun.lwuit.Graphics;
import java.io.IOException;
import java.io.InputStream;


import android.app.Activity;
import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.res.Configuration;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region;
import android.graphics.Typeface;
import android.graphics.Paint.Style;
import android.graphics.Path;
import android.os.Vibrator;
import android.text.InputType;
import android.text.TextPaint;
import android.text.method.DigitsKeyListener;
import android.text.method.PasswordTransformationMethod;
import android.text.method.TextKeyListener;
import android.util.DisplayMetrics;
import android.util.Log;
import android.util.TypedValue;
import android.view.Gravity;
import android.view.View;
import android.view.ViewGroup;
import android.view.Window;
import android.view.inputmethod.InputMethodManager;
import android.widget.EditText;
import android.widget.FrameLayout;
import android.widget.RelativeLayout;
import android.widget.TextView;

import com.sun.lwuit.Component;
import com.sun.lwuit.Display;
import com.sun.lwuit.Font;
import com.sun.lwuit.Form;
import com.sun.lwuit.PeerComponent;
import com.sun.lwuit.TextArea;
import com.sun.lwuit.impl.LWUITImplementation;
import com.sun.lwuit.impl.VirtualKeyboardInterface;
import com.sun.lwuit.plaf.UIManager;
import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.Vector;

public class AndroidImplementation extends LWUITImplementation implements View.OnFocusChangeListener {

    /**
     * make sure these important keys have a negative value when passed
     * to LWUIT or they might be interpreted as characters.
     */
    public static final int DROID_IMPL_KEY_LEFT = -23456;
    public static final int DROID_IMPL_KEY_RIGHT = -23457;
    public static final int DROID_IMPL_KEY_UP = -23458;
    public static final int DROID_IMPL_KEY_DOWN = -23459;
    public static final int DROID_IMPL_KEY_FIRE = -23450;
    public static final int DROID_IMPL_KEY_MENU = -23451;
    public static final int DROID_IMPL_KEY_BACK = -23452;
    public static final int DROID_IMPL_KEY_BACKSPACE = -23453;
    public static final int DROID_IMPL_KEY_CLEAR = -23454;
    public static final int DROID_IMPL_KEY_SYMBOL = -23455;
    static int[] leftSK = new int[]{DROID_IMPL_KEY_MENU};
    private AndroidView myView = null;
    private Paint defaultFont;
    private final char[] tmpchar = new char[1];
    private final RectF tmprectF = new RectF();
    private final Rect tmprect = new Rect();
    private final Path tmppath = new Path();
    protected int defaultFontHeight;
    private boolean keyboardShowing = false;
    private boolean editing = false;
    private int lastSizeChangeW = -1;
    private int lastSizeChangeH = -1;
    private final char[] tmpDrawChar = new char[1];
    private Vibrator v = null;
    private boolean vibrateInitialized = false;
    private Activity activity;
    private RelativeLayout relativeLayout;
    final Vector nativePeers = new Vector();
    private int lastDirectionalKeyEventReceivedByWrapper;

    @Override
    public void init(Object m) {
        if (m instanceof Activity) {
            /**
             * if you call Display.init(..) with a real activity, then this
             * object will be used.
             */
            this.activity = (Activity) m;
        } else {
            /**
             * oterwise assume we use the midlet like baseclass from the
             * javax subpackage.
             */
            this.activity = LWUITActivity.currentActivity;
        }
        // uncomment for fullscreen.
        //activity.getWindow().setFlags(WindowManager.LayoutParams.FLAG_FORCE_NOT_FULLSCREEN,
        //	WindowManager.LayoutParams.FLAG_FORCE_NOT_FULLSCREEN);
        /**
         * translate our default font height depending on the screen density.
         * this is required for new high resolution devices.  otherwise everything
         * looks awfully small.
         *
         * we use our default font height value of 18 and go from there.  i thought
         * about using new Paint().getTextSize() for this value but if some new
         * version of android suddenly returns values already tranlated to the screen
         * then we might end up with too large fonts.  the documentation is not very
         * precise on that.
         */
        final int defaultFontPixelHeight = 18;
        this.defaultFontHeight = this.translatePixelForDPI(defaultFontPixelHeight);


        this.defaultFont = (Paint) this.createFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_MEDIUM);
        Display.getInstance().setTransitionYield(-1);
        this.initSurface();
        /**
         * devices are extemely sensitive so dragging should start
         * a little later than suggested by default implementation.
         */
        this.setDragStartPercentage(6);
        VirtualKeyboardInterface vkb = new AndroidKeyboard(activity, myView);
        Display.getInstance().registerVirtualKeyboard(vkb);
        Display.getInstance().setDefaultVirtualKeyboard(vkb);
    }

    public int translatePixelForDPI(int pixel) {
        return (int) TypedValue.applyDimension(
                TypedValue.COMPLEX_UNIT_SP, pixel,
                activity.getResources().getDisplayMetrics());
    }

    // @Override
    public int getDeviceDensity() {
        DisplayMetrics metrics = new DisplayMetrics();
        activity.getWindowManager().getDefaultDisplay().getMetrics(metrics);
        switch (metrics.densityDpi) {
            case DisplayMetrics.DENSITY_LOW:
                return Display.DENSITY_LOW;
            case DisplayMetrics.DENSITY_HIGH:
                return Display.DENSITY_HIGH;
            default:
                return Display.DENSITY_MEDIUM;
        }
    }

    /**
     * init view. if we are not already on the UI thread, then there is a
     * lot of back and forth to make sure everything is initialized once
     * this method ends.
     */
    void initSurface() {

        final boolean[] initComplete = new boolean[]{false};
        activity.runOnUiThread(new Runnable() {

            /**
             * create views on UI thread!
             */
            @Override
            public void run() {
                relativeLayout = new RelativeLayout(activity);
                relativeLayout.setLayoutParams(new RelativeLayout.LayoutParams(
                        RelativeLayout.LayoutParams.FILL_PARENT,
                        RelativeLayout.LayoutParams.FILL_PARENT));
                relativeLayout.setFocusable(false);
                myView = new AndroidView(activity,
                        AndroidImplementation.this);
                relativeLayout.addView(myView);
                myView.setVisibility(View.VISIBLE);
                activity.setContentView(relativeLayout);
                myView.setOnFocusChangeListener(AndroidImplementation.this);
                myView.requestFocus();

                synchronized (initComplete) {
                    initComplete[0] = true;
                    initComplete.notify();
                }
            }
        });

        synchronized (initComplete) {
            if (!initComplete[0]) {
                try {
                    initComplete.wait();
                } catch (Exception ignored) {
                    ;
                }
            }
        }
    }

    @Override
    public void confirmControlView() {
        activity.runOnUiThread(new Runnable() {

            public void run() {
                myView.setVisibility(View.VISIBLE);
            }
        });
    }

    protected boolean editInProgress() {
        return this.editing;
    }

    public void hideNotifyPublic() {
        super.hideNotify();
    }

    public void showNotifyPublic() {
        super.showNotify();
    }

    @Override
    public boolean isMinimized() {
        return myView.getVisibility() != View.VISIBLE;
    }

    @Override
    public boolean minimizeApplication() {
        activity.runOnUiThread(new Runnable() {

            public void run() {
                myView.setVisibility(View.INVISIBLE);
            }
        });
        return true;
    }

    @Override
    public void restoreMinimizedApplication() {
        activity.runOnUiThread(new Runnable() {

            public void run() {
                myView.setVisibility(View.VISIBLE);
            }
        });
    }

    @Override
    public void editString(final Component cmp, int maxSize, final int constraint,
            final String text) {

        /**
         * softkeyboard should open on a short click on the native text area.
         */
        final boolean[] editingComplete = new boolean[]{false};
        this.editing = true;

        activity.runOnUiThread(new Runnable() {

            @Override
            public void run() {
                UIManager m = UIManager.getInstance();
                final FrameLayout frameLayout = new FrameLayout(
                        activity);
                final EditText editText = new EditText(
                        activity);
                editText.setGravity(Gravity.CENTER | Gravity.FILL_HORIZONTAL);
                editText.setSingleLine(true);

                switch (constraint) {
                    case TextArea.NUMERIC:
                        editText.setKeyListener(DigitsKeyListener.getInstance(true, false));
                        break;
                    case TextArea.DECIMAL:
                        editText.setKeyListener(DigitsKeyListener.getInstance(true, true));
                        break;
                    case TextArea.PASSWORD:
                        // turns out the setKeyListener(..) and setTransformationMethod(..) methods alone
                        // would not disable the input hints, causing the device to attempt to save passwords.
                        // not good!  using setInputType(..) works.
                        editText.setInputType(InputType.TYPE_CLASS_TEXT | InputType.TYPE_TEXT_VARIATION_PASSWORD);
                        //editText.setKeyListener(TextKeyListener.getInstance(false, TextKeyListener.Capitalize.NONE));
                        //editText.setTransformationMethod(PasswordTransformationMethod.getInstance());
                        break;
                    case TextArea.INITIAL_CAPS_WORD:
                        editText.setKeyListener(TextKeyListener.getInstance(false, TextKeyListener.Capitalize.WORDS));
                        break;
                    case TextArea.INITIAL_CAPS_SENTENCE:
                        editText.setKeyListener(TextKeyListener.getInstance(false, TextKeyListener.Capitalize.SENTENCES));
                        break;
                    default:
                        editText.setKeyListener(TextKeyListener.getInstance(false, TextKeyListener.Capitalize.NONE));
                        break;
                }

                frameLayout.addView(editText, new FrameLayout.LayoutParams(
                        FrameLayout.LayoutParams.FILL_PARENT,
                        FrameLayout.LayoutParams.WRAP_CONTENT));
                editText.setText(text == null ? "" : text);
                AlertDialog.Builder bob = new AlertDialog.Builder(
                        activity);
                frameLayout.setMinimumWidth(Math.min(
                        AndroidImplementation.this.translatePixelForDPI(250),
                        Math.min(AndroidImplementation.this.getDisplayHeight(),
                        AndroidImplementation.this.getDisplayWidth())));
                bob.setView(frameLayout);
                bob.setTitle("");
                bob.setPositiveButton(m.localize("ok", "OK"),
                        new DialogInterface.OnClickListener() {

                            @Override
                            public void onClick(DialogInterface d, int which) {
                                Display.getInstance().onEditingComplete(cmp,
                                        editText.getText().toString());
                                d.dismiss();
                            }
                        });
                bob.setNegativeButton(m.localize("cancel", "Cancel"),
                        new DialogInterface.OnClickListener() {

                            @Override
                            public void onClick(DialogInterface d, int which) {
                                d.dismiss();
                            }
                        });
                AlertDialog editDialog = bob.create();
                editDialog.setOnDismissListener(new DialogInterface.OnDismissListener() {

                    public void onDismiss(DialogInterface arg0) {
                        synchronized (editingComplete) {
                            editingComplete[0] = true;
                            editingComplete.notify();
                        }
                    }
                });
                editDialog.show();
            }
        });

        synchronized (editingComplete) {
            if (!editingComplete[0]) {
                try {
                    editingComplete.wait();
                } catch (Exception ignored) {
                    ;
                }
            }
        }

        this.editing = false;

        if (this.lastSizeChangeH != -1 && this.lastSizeChangeW != -1) {
            this.myView.handleSizeChange(this.lastSizeChangeW, this.lastSizeChangeH);
        }
        this.lastSizeChangeH = -1;
        this.lastSizeChangeW = -1;
    }

    protected void setLastSizeChangedWH(int w, int h) {
        this.lastSizeChangeW = w;
        this.lastSizeChangeH = h;
    }

    @Override
    public boolean handleEDTException(final Throwable err) {

        final boolean[] messageComplete = new boolean[]{false};

        Log.e("LWUIT", "Err on EDT", err);

        activity.runOnUiThread(new Runnable() {

            @Override
            public void run() {
                UIManager m = UIManager.getInstance();
                final FrameLayout frameLayout = new FrameLayout(
                        activity);
                final TextView textView = new TextView(
                        activity);
                textView.setGravity(Gravity.CENTER);
                frameLayout.addView(textView, new FrameLayout.LayoutParams(
                        FrameLayout.LayoutParams.FILL_PARENT,
                        FrameLayout.LayoutParams.WRAP_CONTENT));
                textView.setText("An internal application error occurred: " + err.toString());
                AlertDialog.Builder bob = new AlertDialog.Builder(
                        activity);
                bob.setView(frameLayout);
                bob.setTitle("");
                bob.setPositiveButton(m.localize("ok", "OK"),
                        new DialogInterface.OnClickListener() {

                            @Override
                            public void onClick(DialogInterface d, int which) {
                                d.dismiss();
                                synchronized (messageComplete) {
                                    messageComplete[0] = true;
                                    messageComplete.notify();
                                }
                            }
                        });
                AlertDialog editDialog = bob.create();
                editDialog.show();
            }
        });

        synchronized (messageComplete) {
            if (messageComplete[0]) {
                return true;
            }
            try {
                messageComplete.wait();
            } catch (Exception ignored) {
                ;
            }
        }
        return true;
    }

    //@Override
    public InputStream getResourceAsStream(Class cls, String resource) {
        try {
            if (resource.startsWith("/")) {
                resource = resource.substring(1);
            }
            return activity.getAssets().open(resource);
        } catch (IOException ex) {
            Log.e("LWUIT", "Failed to load resource: " + resource, ex);
            return null;
        }
    }

    @Override
    protected void pointerPressed(int x, int y) {
        super.pointerPressed(x, y);
    }

    @Override
    protected void pointerReleased(int x, int y) {
        super.pointerReleased(x, y);
    }

    @Override
    protected void pointerDragged(int x, int y) {
        super.pointerDragged(x, y);
        /**
         * devices fire a crazy amount of drag events.  slowing this down
         * a little. uncomment this if you feel it helps.
         * http://www.youtube.com/watch?v=U4Bk5rmIpic
         * at 0:55:35
         */
        //try {
        //    Thread.sleep(20);
        //} catch (Exception ignored) {
        //    ;
        //}
    }

    @Override
    protected int getDragAutoActivationThreshold() {
        return 1000000;
    }

    @Override
    public void flushGraphics() {
        myView.flushGraphics();
    }

    @Override
    public void flushGraphics(int x, int y, int width, int height) {
        this.tmprect.set(x, y, x + width, y + height);
        myView.flushGraphics(this.tmprect);
    }

    @Override
    public int charWidth(Object nativeFont, char ch) {
        this.tmpchar[0] = ch;
        float w = (nativeFont == null ? this.defaultFont
                : (Paint) nativeFont).measureText(this.tmpchar, 0, 1);
        if (w - (int) w > 0) {
            return (int) (w + 1);
        }
        return (int) w;
    }

    @Override
    public int charsWidth(Object nativeFont, char[] ch, int offset, int length) {
        float w = (nativeFont == null ? this.defaultFont
                : (Paint) nativeFont).measureText(ch, offset, length);
        if (w - (int) w > 0) {
            return (int) (w + 1);
        }
        return (int) w;
    }

    @Override
    public int stringWidth(Object nativeFont, String str) {
        float w = (nativeFont == null ? this.defaultFont
                : (Paint) nativeFont).measureText(str);
        if (w - (int) w > 0) {
            return (int) (w + 1);
        }
        return (int) w;
    }

    @Override
    public void setNativeFont(Object graphics, Object font) {
        if (font == null) {
            font = this.defaultFont;
        }
        ((AndroidGraphics) graphics).setFont((Paint) font);
    }

    @Override
    public int getHeight(Object nativeFont) {
        Paint font = (nativeFont == null ? this.defaultFont
                : (Paint) nativeFont);
        return font.getFontMetricsInt(font.getFontMetricsInt());
    }

    @Override
    public int getSize(Object nativeFont) {
        int height = this.getHeight(nativeFont);
        if (height < defaultFontHeight) {
            return Font.SIZE_SMALL;
        }
        if (height > defaultFontHeight) {
            return Font.SIZE_LARGE;
        }
        return Font.SIZE_MEDIUM;
    }

    @Override
    public int getStyle(Object nativeFont) {
        // http://www.java.net/forum/topic/mobile-embedded/lwuit/android-implementation-font-issue
        if (nativeFont instanceof Paint) {
            Paint new_name = (Paint) nativeFont;
            return new_name.getTypeface().getStyle();
        }
        return 0;
    }

    @Override
    public Object createFont(int face, int style, int size) {
        Paint font = new TextPaint();
        font.setAntiAlias(true);
        Typeface typeface = null;
        switch (face) {
            case Font.FACE_MONOSPACE:
                typeface = Typeface.MONOSPACE;
                break;
            default:
                typeface = Typeface.DEFAULT;
                break;
        }

        int fontstyle = Typeface.NORMAL;
        if ((style & Font.STYLE_BOLD) != 0) {
            fontstyle |= Typeface.BOLD;
        }
        if ((style & Font.STYLE_ITALIC) != 0) {
            fontstyle |= Typeface.ITALIC;
        }


        int height = this.defaultFontHeight;
        int diff = height / 3;

        switch (size) {
            case Font.SIZE_SMALL:
                height -= diff;
                break;
            case Font.SIZE_LARGE:
                height += diff;
                break;
        }

        font.setTypeface(Typeface.create(typeface, fontstyle));
        font.setUnderlineText((style & Font.STYLE_UNDERLINED) != 0);
        font.setTextSize(height);
        return font;

    }

    /**
     * Loads a native font based on a lookup for a font name and attributes. Font lookup
     * values can be separated by commas and thus allow fallback if the primary font
     * isn't supported by the platform.
     *
     * @param lookup string describing the font
     * @return the native font object
     */
    public Object loadNativeFont(String lookup) {
        Paint font = new TextPaint();
        font.setAntiAlias(true);
        int typeface = Typeface.NORMAL;
        String familyName = lookup.substring(0, lookup.indexOf("-"));
        String style = lookup.substring(lookup.indexOf("-") + 1, lookup.lastIndexOf("-"));
        String size = lookup.substring(lookup.lastIndexOf("-") + 1, lookup.length());

        if (style.equals("bolditalic")) {
            typeface = Typeface.BOLD_ITALIC;
        } else if (style.equals("italic")) {
            typeface = Typeface.ITALIC;
        } else if (style.equals("bold")) {
            typeface = Typeface.BOLD;
        }
        font.setTypeface(Typeface.create(familyName, typeface));
        font.setTextSize(Integer.parseInt(size));
        return font;
    }

    /**
     * Indicates whether loading a font by a string is supported by the platform
     *
     * @return true if the platform supports font lookup
     */
    public boolean isLookupFontSupported() {
        return true;
    }

    @Override
    public boolean isAntiAliasedTextSupported() {
        return true;
    }

    @Override
    public void setAntiAliasedText(Object graphics, boolean a) {
        ((AndroidGraphics) graphics).getFont().setAntiAlias(a);
    }

    @Override
    public Object getDefaultFont() {
        TextPaint paint = new TextPaint();
        paint.set(this.defaultFont);
        return paint;
    }

    @Override
    public Object getNativeGraphics() {
        return this.myView.getGraphics();
    }

    @Override
    public Object getNativeGraphics(Object image) {
        return new AndroidGraphics(this, new Canvas((Bitmap) image));
    }

    @Override
    public void getRGB(Object nativeImage, int[] arr, int offset, int x, int y,
            int width, int height) {
        ((Bitmap) nativeImage).getPixels(arr, offset, width, x, y, width,
                height);
    }

    @Override
    public Object createImage(String path) throws IOException {
        InputStream in = this.getResourceAsStream(null, path);
        if (in == null) {
            throw new IOException("Resource not found. " + path);
        }
        try {
            return this.createImage(in);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (Exception ignored) {
                    ;
                }
            }
        }
    }

    @Override
    public Object createImage(InputStream i) throws IOException {
        BitmapFactory.Options opts = new BitmapFactory.Options();
        try {
            BitmapFactory.Options.class.getField("inPurgeable").set(opts, true);
        } catch (Exception e) {
            // inPurgeable not supported
            // http://www.droidnova.com/2d-sprite-animation-in-android-addendum,505.html
        }
        return BitmapFactory.decodeStream(i, null, opts);
    }

    @Override
    public Object createImage(byte[] bytes, int offset, int len) {
        BitmapFactory.Options opts = new BitmapFactory.Options();
        try {
            BitmapFactory.Options.class.getField("inPurgeable").set(opts, true);
        } catch (Exception e) {
            // inPurgeable not supported
            // http://www.droidnova.com/2d-sprite-animation-in-android-addendum,505.html
        }
        return BitmapFactory.decodeByteArray(bytes, offset, len, opts);
    }

    @Override
    public Object createImage(int[] rgb, int width, int height) {
        return Bitmap.createBitmap(rgb, width, height, Bitmap.Config.ARGB_8888);
    }

    @Override
    public boolean isAlphaMutableImageSupported() {
        return true;
    }

    @Override
    public Object scale(Object nativeImage, int width, int height) {
        return Bitmap.createScaledBitmap((Bitmap) nativeImage, width, height,
                false);
    }

    @Override
    public Object rotate(Object image, int degrees) {
        Matrix matrix = new Matrix();
        matrix.postRotate(degrees);
        return Bitmap.createBitmap((Bitmap) image, 0, 0, ((Bitmap) image).getWidth(), ((Bitmap) image).getHeight(), matrix, true);
    }

    @Override
    public boolean isRotationDrawingSupported() {
        return false;
    }

    @Override
    public Object createMutableImage(int width, int height, int fillColor) {
        Bitmap bitmap = Bitmap.createBitmap(width, height,
                Bitmap.Config.ARGB_8888);
        AndroidGraphics graphics = (AndroidGraphics) this.getNativeGraphics(bitmap);
        this.setColor(graphics, fillColor);
        this.fillRect(graphics, 0, 0, width, height);
        return bitmap;
    }

    @Override
    public int getImageHeight(Object i) {
        return ((Bitmap) i).getHeight();
    }

    @Override
    public int getImageWidth(Object i) {
        return ((Bitmap) i).getWidth();
    }

    @Override
    public void drawImage(Object graphics, Object img, int x, int y) {
        ((AndroidGraphics) graphics).getCanvas().drawBitmap((Bitmap) img, x, y, null);
    }

    @Override
    public void drawLine(Object graphics, int x1, int y1, int x2, int y2) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.STROKE);
        ((AndroidGraphics) graphics).getCanvas().drawLine(x1, y1, x2, y2,
                ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public boolean isAntiAliasingSupported() {
        return true;
    }

    @Override
    public void setAntiAliased(Object graphics, boolean a) {
        ((AndroidGraphics) graphics).getPaint().setAntiAlias(a);
    }

    @Override
    public void drawPolygon(Object graphics, int[] xPoints, int[] yPoints, int nPoints) {
        if (nPoints <= 1) {
            return;
        }
        this.tmppath.rewind();
        this.tmppath.moveTo(xPoints[0], yPoints[0]);
        for (int i = 1; i < nPoints; i++) {
            this.tmppath.lineTo(xPoints[i], yPoints[i]);
        }
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.STROKE);
        ((AndroidGraphics) graphics).getCanvas().drawPath(this.tmppath, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void fillPolygon(Object graphics, int[] xPoints, int[] yPoints, int nPoints) {
        if (nPoints <= 1) {
            return;
        }
        this.tmppath.rewind();
        this.tmppath.moveTo(xPoints[0], yPoints[0]);
        for (int i = 1; i < nPoints; i++) {
            this.tmppath.lineTo(xPoints[i], yPoints[i]);
        }
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.FILL);
        ((AndroidGraphics) graphics).getCanvas().drawPath(this.tmppath, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void drawRGB(Object graphics, int[] rgbData, int offset, int x,
            int y, int w, int h, boolean processAlpha) {
        //Bitmap tmp = Bitmap.createBitmap(rgbData, w, h, Bitmap.Config.ARGB_8888);
        //((AndroidGraphics) graphics).drawBitmap(tmp, x, y, null);
        ((AndroidGraphics) graphics).getCanvas().drawBitmap(rgbData, offset, w, x, y, w, h,
                processAlpha, null);
    }

    @Override
    public void drawRect(Object graphics, int x, int y, int width, int height) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.STROKE);
        ((AndroidGraphics) graphics).getCanvas().drawRect(x, y, x + width, y + height,
                ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void drawRoundRect(Object graphics, int x, int y, int width,
            int height, int arcWidth, int arcHeight) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.STROKE);
        this.tmprectF.set(x, y, x + width, y + height);
        ((AndroidGraphics) graphics).getCanvas().drawRoundRect(this.tmprectF, arcWidth,
                arcHeight, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void drawString(Object graphics, String str, int x, int y) {
        ((AndroidGraphics) graphics).getCanvas().drawText(str, x, y - ((AndroidGraphics) graphics).getFont().getFontMetricsInt().ascent,
                ((AndroidGraphics) graphics).getFont());
    }

    /**
     *  the next two methods are not used yet and are part of a potential performance enhancement.
     *  see https://lwuit.dev.java.net/issues/show_bug.cgi?id=218
     *  for details.
     */
    //@Override
    public void drawChar(Object graphics, char c, int x, int y) {
        tmpDrawChar[0] = c;
        ((AndroidGraphics) graphics).getCanvas().drawText(tmpDrawChar, 0, 1, x, y - ((AndroidGraphics) graphics).getFont().getFontMetricsInt().ascent,
                ((AndroidGraphics) graphics).getFont());
    }

    //@Override
    public void drawChars(Object graphics, char[] c, int offset, int length, int x, int y) {
        ((AndroidGraphics) graphics).getCanvas().drawText(c, offset, length, x, y - ((AndroidGraphics) graphics).getFont().getFontMetricsInt().ascent,
                ((AndroidGraphics) graphics).getFont());
    }

    @Override
    public void drawArc(Object graphics, int x, int y, int width, int height,
            int startAngle, int arcAngle) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.STROKE);
        this.tmprectF.set(x, y, x + width, y + height);
        ((AndroidGraphics) graphics).getCanvas().drawArc(this.tmprectF, startAngle,
                arcAngle, false, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void fillArc(Object graphics, int x, int y, int width, int height,
            int startAngle, int arcAngle) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.FILL);
        this.tmprectF.set(x, y, x + width, y + height);
        ((AndroidGraphics) graphics).getCanvas().drawArc(this.tmprectF, startAngle,
                arcAngle, false, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void fillRect(Object graphics, int x, int y, int width, int height) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.FILL);
        ((AndroidGraphics) graphics).getCanvas().drawRect(x, y, x + width, y + height,
                ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public void fillRoundRect(Object graphics, int x, int y, int width,
            int height, int arcWidth, int arcHeight) {
        ((AndroidGraphics) graphics).getPaint().setStyle(Style.FILL);
        this.tmprectF.set(x, y, x + width, y + height);
        ((AndroidGraphics) graphics).getCanvas().drawRoundRect(this.tmprectF, arcWidth,
                arcHeight, ((AndroidGraphics) graphics).getPaint());
    }

    @Override
    public int getAlpha(Object graphics) {
        return ((AndroidGraphics) graphics).getPaint().getAlpha();
    }

    @Override
    public void setAlpha(Object graphics, int alpha) {
        ((AndroidGraphics) graphics).getPaint().setAlpha(alpha);
    }

    @Override
    public boolean isAlphaGlobal() {
        return true;
    }

    @Override
    public void setColor(Object graphics, int RGB) {
        ((AndroidGraphics) graphics).setColor(0xff000000 | RGB);
    }

    @Override
    public int getBackKeyCode() {
        return DROID_IMPL_KEY_BACK;
    }

    @Override
    public int getBackspaceKeyCode() {
        return DROID_IMPL_KEY_BACKSPACE;
    }

    @Override
    public int getClearKeyCode() {
        return DROID_IMPL_KEY_CLEAR;
    }

    @Override
    public int getClipHeight(Object graphics) {
        ((AndroidGraphics) graphics).getCanvas().getClipBounds(this.tmprect);
        return this.tmprect.height();
    }

    @Override
    public int getClipWidth(Object graphics) {
        ((AndroidGraphics) graphics).getCanvas().getClipBounds(this.tmprect);
        return this.tmprect.width();
    }

    @Override
    public int getClipX(Object graphics) {
        ((AndroidGraphics) graphics).getCanvas().getClipBounds(this.tmprect);
        return this.tmprect.left;
    }

    @Override
    public int getClipY(Object graphics) {
        ((AndroidGraphics) graphics).getCanvas().getClipBounds(this.tmprect);
        return this.tmprect.top;
    }

    @Override
    public void setClip(Object graphics, int x, int y, int width, int height) {
        ((AndroidGraphics) graphics).getCanvas().clipRect(x, y, x + width, y + height, Region.Op.REPLACE);
    }

    @Override
    public void clipRect(Object graphics, int x, int y, int width, int height) {
        ((AndroidGraphics) graphics).getCanvas().clipRect(x, y, x + width, y + height);
    }

    @Override
    public int getColor(Object graphics) {
        return ((AndroidGraphics) graphics).getPaint().getColor();
    }

    @Override
    public int getDisplayHeight() {
        return this.myView.getViewHeight();
    }

    @Override
    public int getDisplayWidth() {
        return this.myView.getViewWidth();
    }

    @Override
    public int getGameAction(int keyCode) {
        switch (keyCode) {
            case DROID_IMPL_KEY_DOWN:
                return Display.GAME_DOWN;
            case DROID_IMPL_KEY_UP:
                return Display.GAME_UP;
            case DROID_IMPL_KEY_LEFT:
                return Display.GAME_LEFT;
            case DROID_IMPL_KEY_RIGHT:
                return Display.GAME_RIGHT;
            case DROID_IMPL_KEY_FIRE:
                return Display.GAME_FIRE;
            default:
                return 0;
        }
    }

    @Override
    public int getKeyCode(int gameAction) {
        switch (gameAction) {
            case Display.GAME_DOWN:
                return DROID_IMPL_KEY_DOWN;
            case Display.GAME_UP:
                return DROID_IMPL_KEY_UP;
            case Display.GAME_LEFT:
                return DROID_IMPL_KEY_LEFT;
            case Display.GAME_RIGHT:
                return DROID_IMPL_KEY_RIGHT;
            case Display.GAME_FIRE:
                return DROID_IMPL_KEY_FIRE;
            default:
                return 0;
        }
    }

    @Override
    public int[] getSoftkeyCode(int index) {
        if (index == 0) {
            return leftSK;
        }
        return null;
    }

    @Override
    public int getSoftkeyCount() {
        /**
         * one menu button only.  we may have to stuff some code here
         * as soon as there are devices that no longer have only a single
         * menu button.
         */
        return 1;
    }

    @Override
    public void vibrate(int duration) {
        if (!this.vibrateInitialized) {
            try {
                v = (Vibrator) activity.getSystemService(Context.VIBRATOR_SERVICE);
            } catch (Throwable e) {
                Log.e("LWUIT", "problem with virbrator(0)", e);
            } finally {
                this.vibrateInitialized = true;
            }
        }
        if (v != null) {
            try {
                v.vibrate(duration);
            } catch (Throwable e) {
                Log.e("LWUIT", "problem with virbrator(1)", e);
            }
        }
    }

    @Override
    public boolean isTouchDevice() {
        Configuration c = myView.getResources().getConfiguration();
        return c.touchscreen != Configuration.TOUCHSCREEN_NOTOUCH;
    }

    @Override
    public boolean hasPendingPaints() {
        //if the view is not visible make sure the edt won't wait.
        if (myView.getVisibility() != View.VISIBLE) {
            return true;
        } else {
            return super.hasPendingPaints();
        }
    }

    public void revalidate() {
        myView.setVisibility(View.VISIBLE);
        getCurrentForm().revalidate();
        flushGraphics();
    }

    @Override
    public int getKeyboardType() {
        if (Display.getInstance().getDefaultVirtualKeyboard().isVirtualKeyboardShowing()) {
            return Display.KEYBOARD_TYPE_VIRTUAL;
        }
        /**
         * can we detect this?  but even if we could i think
         * it is best to have this fixed to qwerty.  we pass unicode
         * values to lwuit in any case.  check AndroidView.onKeyUpDown()
         * method.  and read comment below.
         */
        return Display.KEYBOARD_TYPE_QWERTY;
        /**
         * some info from the MIDP docs about keycodes:
         *
         *  "Applications receive keystroke events in which the individual keys are named within a space of key codes.
         * Every key for which events are reported to MIDP applications is assigned a key code. The key code values are
         * unique for each hardware key unless two keys are obvious synonyms for each other. MIDP defines the following
         * key codes: KEY_NUM0, KEY_NUM1, KEY_NUM2, KEY_NUM3, KEY_NUM4, KEY_NUM5, KEY_NUM6, KEY_NUM7, KEY_NUM8, KEY_NUM9,
         * KEY_STAR, and KEY_POUND. (These key codes correspond to keys on a ITU-T standard telephone keypad.) Other
         * keys may be present on the keyboard, and they will generally have key codes distinct from those list above.
         * In order to guarantee portability, applications should use only the standard key codes.
         *
         * The standard key codes' values are equal to the Unicode encoding for the character that represents the key.
         * If the device includes any other keys that have an obvious correspondence to a Unicode character, their key
         * code values should equal the Unicode encoding for that character. For keys that have no corresponding Unicode
         * character, the implementation must use negative values. Zero is defined to be an invalid key code."
         *
         * Because the MIDP implementation is our reference and that implementation does not interpret the given keycodes
         * we behave alike and pass on the unicode values.
         */
    }

    // @Override
    public Object createSoftWeakRef(Object o) {
        return new SoftReference(o);
    }

    // @Override
    public Object extractHardRef(Object o) {
        return o == null ? null : ((SoftReference) o).get();
    }

    /**
     * @deprecated
     */
    public boolean isVirtualKeyboardShowing() {
        return this.keyboardShowing;
    }

    /**
     * @deprecated
     */
    public boolean isVirtualKeyboardShowingSupported() {
        return this.isTouchDevice();
    }

    /**
     * @deprecated
     */
    public void setShowVirtualKeyboard(boolean show) {

        /**
         * to properly handle the virtual keyboard i see no other way than implementing
         * a customized input method service.  crap.
         *
         * http://www.mail-archive.com/android-developers@googlegroups.com/msg37058.html
         * http://groups.google.com/group/android-developers/browse_thread/thread/0da0d2eff53a7b55
         * https://groups.google.com/group/android-developers/browse_thread/thread/e23040e899eb0a0a
         * http://www.go-android.de/externe-news/creating-input-method
         */
        if (this.keyboardShowing != show) {
            // http://android-developers.blogspot.com/2009/04/updating-applications-for-on-screen.html
            InputMethodManager manager = (InputMethodManager) activity.getSystemService(Context.INPUT_METHOD_SERVICE);
            manager.toggleSoftInputFromWindow(myView.getWindowToken(), 0, 0);
            this.keyboardShowing = show;

            /*
             * suggested alternative from the forums, could not test it yet.
            InputMethodManager manager = (InputMethodManager) activity.getSystemService(Context.INPUT_METHOD_SERVICE);
            if (show) {
            manager.showSoftInput(myView, 0);
            } else {
            manager.hideSoftInputFromWindow(myView.getWindowToken(), 0);
            }
            this.keyboardShowing = show;
             *
             */
        }
    }

    /**
     * @deprecated
     */
    protected void clearVirtualKeyboard() {
        this.keyboardShowing = false;
    }

    // @Override
    public void showNativeScreen(Object nativeFullScreenPeer) {
        if (!(nativeFullScreenPeer instanceof View)) {
            throw new IllegalArgumentException(nativeFullScreenPeer.getClass().getName());
        }
        // todo
    }

    /**
     * Places the following commands on the native menu system
     *
     * @param commands the LWUIT commands to use
     */
    // @Override
    public void setNativeCommands(Vector commands) {
        // todo
    }

    // @Override
    public PeerComponent createNativePeer(Object nativeComponent) {
        if (!(nativeComponent instanceof View)) {
            throw new IllegalArgumentException(nativeComponent.getClass().getName());
        }
        return new AndroidPeer((View) nativeComponent);
    }

    private void blockNativeFocusAll(boolean block) {
        synchronized (this.nativePeers) {
            final int size = this.nativePeers.size();
            for (int i = 0; i < size; i++) {
                AndroidPeer next = (AndroidPeer) this.nativePeers.get(i);
                next.blockNativeFocus(block);
            }
        }
    }

    public void onFocusChange(View view, boolean bln) {

        if (bln) {
            /**
             * whenever the base view receives focus we automatically block possible
             * native subviews from gaining focus.
             */
            blockNativeFocusAll(true);
            if (this.lastDirectionalKeyEventReceivedByWrapper != 0) {
                /**
                 * because we also consume any key event in the OnKeyListener of the native wrappers,
                 * we have to simulate key events to make LWUIT move the focus to the next component.
                 */
                if (!myView.isInTouchMode()) {
                    switch (lastDirectionalKeyEventReceivedByWrapper) {
                        case AndroidImplementation.DROID_IMPL_KEY_LEFT:
                        case AndroidImplementation.DROID_IMPL_KEY_RIGHT:
                        case AndroidImplementation.DROID_IMPL_KEY_UP:
                        case AndroidImplementation.DROID_IMPL_KEY_DOWN:
                            Log.d("LWUIT", "sending keycode: " + lastDirectionalKeyEventReceivedByWrapper);
                            Display.getInstance().keyPressed(lastDirectionalKeyEventReceivedByWrapper);
                            Display.getInstance().keyReleased(lastDirectionalKeyEventReceivedByWrapper);
                            break;
                        default:
                            Log.d("LWUIT", "unexpected keycode: " + lastDirectionalKeyEventReceivedByWrapper);
                            break;
                    }
                } else {
                    Log.d("LWUIT", "base view gained focus but no key event to process.");
                }
                lastDirectionalKeyEventReceivedByWrapper = 0;
            }
        }

    }

    /**
     * wrapper component that capsules a native view object in a LWUIT component. this
     * involves A LOT of back and forth between the LWUIT EDT and the Android UI thread.
     *
     *
     * To use it you would:
     *
     * 1) create your native Android view(s). Make sure to work on the Android UI thread when constructing
     *    and modifying them.
     * 2) create a LWUIT peer component by calling:
     *
     *         com.sun.lwuit.PeerComponent.create(myAndroidView);
     *
     * 3) currently the view's size is not automatically calculated from the native view. so you should
     *    set the preferred size of the LWUIT component manually.
     *
     *
     */
    class AndroidPeer extends PeerComponent {

        private View v;
        AndroidRelativeLayout layoutWrapper = null;

        public AndroidPeer(View vv) {
            super(vv);
            this.v = vv;
        }

        @Override
        public void setVisible(boolean visible) {
            // EDT
            super.setVisible(visible);
            this.doSetVisibility(visible);
        }

        void doSetVisibility(final boolean visible) {
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    v.setVisibility(visible ? View.VISIBLE : View.INVISIBLE);
                    if (visible) {
                        v.bringToFront();
                    }
                }
            });
        }

        @Override
        protected void deinitialize() {
            // EDT
            super.deinitialize();
            synchronized (nativePeers) {
                nativePeers.remove(this);
            }
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    if (layoutWrapper != null) {
                        AndroidImplementation.this.relativeLayout.removeView(layoutWrapper);
                    }
                }
            });
        }

        @Override
        protected void initComponent() {
            // EDT
            super.initComponent();
            synchronized (nativePeers) {
                nativePeers.add(this);
            }
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    if (layoutWrapper == null) {
                        /**
                         * wrap the native item in a layout that we can move around on the surface view
                         * as we like.
                         */
                        layoutWrapper = new AndroidRelativeLayout(activity, AndroidPeer.this, v);
                        v.setFocusable(AndroidPeer.this.isFocusable());
                        v.setFocusableInTouchMode(AndroidPeer.this.isFocusable());
                        ArrayList<View> viewList = new ArrayList<View>();
                        viewList.add(layoutWrapper);
                        v.addFocusables(viewList, View.FOCUS_DOWN);
                        v.addFocusables(viewList, View.FOCUS_UP);
                        v.addFocusables(viewList, View.FOCUS_LEFT);
                        v.addFocusables(viewList, View.FOCUS_RIGHT);
                        if (v.isFocusable() || v.isFocusableInTouchMode()) {
                            if (AndroidPeer.super.hasFocus()) {
                                AndroidImplementation.this.blockNativeFocusAll(true);
                                blockNativeFocus(false);
                                v.requestFocus();
                            } else {
                                blockNativeFocus(true);
                            }
                        } else {
                            blockNativeFocus(true);
                        }
                        layoutWrapper.setOnKeyListener(new View.OnKeyListener() {

                            public boolean onKey(View view, int i, KeyEvent ke) {
                                lastDirectionalKeyEventReceivedByWrapper = AndroidView.internalKeyCodeTranslate(ke.getKeyCode());

                                // move focus back to base view.
                                AndroidImplementation.this.myView.requestFocus();

                                /**
                                 * if the wrapper has focus, then only because the wrapped
                                 * native component just lost focus. we consume whatever key
                                 * events we receive, just to make sure no half press/release
                                 * sequence reaches the base view (and therefore LWUIT).
                                 */
                                return true;
                            }
                        });
                        layoutWrapper.setOnFocusChangeListener(new View.OnFocusChangeListener() {

                            public void onFocusChange(View view, boolean bln) {
                                Log.d("LWUIT", "on focus change. " + view.toString() + " focus:" + bln + " touchmode: "
                                        + v.isInTouchMode());
                            }
                        });
                    }
                    AndroidImplementation.this.relativeLayout.addView(layoutWrapper);
                }
            });
        }

        @Override
        protected void onPositionSizeChange() {

            // called by LWUIT EDT to position the native component.

            activity.runOnUiThread(new Runnable() {

                public void run() {
                    if (layoutWrapper != null) {
                        if (v.getVisibility() == View.VISIBLE) {

                            RelativeLayout.LayoutParams layoutParams = layoutWrapper.createMyLayoutParams(
                                    AndroidPeer.this.getAbsoluteX(),
                                    AndroidPeer.this.getAbsoluteY(),
                                    AndroidPeer.this.getWidth(),
                                    AndroidPeer.this.getHeight());
                            layoutWrapper.setLayoutParams(layoutParams);
                            AndroidImplementation.this.relativeLayout.requestLayout();

                            //Log.d("LWUIT", "new peer size: " + layoutParams.leftMargin + " " + layoutParams.rightMargin
                            //      + " " + layoutParams.width + " " + layoutParams.height);
                        }
                    }
                }
            });
        }

        void blockNativeFocus(boolean block) {
            layoutWrapper.setDescendantFocusability(block
                    ? ViewGroup.FOCUS_BLOCK_DESCENDANTS : ViewGroup.FOCUS_AFTER_DESCENDANTS);
        }

        @Override
        public boolean isFocusable() {
            // EDT
            if (v != null) {
                return v.isFocusableInTouchMode() || v.isFocusable();
            } else {
                return super.isFocusable();
            }
        }

        @Override
        public void setFocusable(final boolean focusable) {
            // EDT
            super.setFocusable(focusable);
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    v.setFocusable(focusable);
                    v.setFocusableInTouchMode(focusable);
                }
            });
        }

        @Override
        protected void focusGained() {
            Log.d("LWUIT", "native focus gain");
            // EDT
            super.focusGained();
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    // allow this one to gain focus
                    blockNativeFocus(false);
                    if (v.isInTouchMode()) {
                        v.requestFocusFromTouch();
                    } else {
                        v.requestFocus();
                    }
                }
            });
        }

        @Override
        protected void focusLost() {
            Log.d("LWUIT", "native focus loss");
            // EDT
            super.focusLost();
            activity.runOnUiThread(new Runnable() {

                public void run() {
                    // request focus of the wrapper. that will trigger the
                    // android focus listener and move focus back to the
                    // base view.
                    layoutWrapper.requestFocus();
                }
            });
        }
    }

    /**
     * inner class that wrapps the native components.
     * this is a useful thingy to handle focus stuff and
     * buffering.
     */
    class AndroidRelativeLayout extends RelativeLayout {

        AndroidPeer peer;
        Bitmap cache;

        public AndroidRelativeLayout(Context activity, AndroidPeer peer, View v) {
            super(activity);
            this.peer = peer;
            this.setLayoutParams(createMyLayoutParams(peer.getAbsoluteX(), peer.getAbsoluteY(),
                    peer.getWidth(), peer.getHeight()));
            this.addView(v, new RelativeLayout.LayoutParams(
                    RelativeLayout.LayoutParams.FILL_PARENT,
                    RelativeLayout.LayoutParams.FILL_PARENT));
            this.setDrawingCacheEnabled(false);
            this.setAlwaysDrawnWithCacheEnabled(false);
            this.setFocusable(true);
            this.setFocusableInTouchMode(false);
            this.setDescendantFocusability(ViewGroup.FOCUS_AFTER_DESCENDANTS);
        }

        /**
         * create a layout parameter object that holds the native component's position.
         * @return
         */
        private RelativeLayout.LayoutParams createMyLayoutParams(int x, int y, int width, int height) {
            RelativeLayout.LayoutParams layoutParams = new RelativeLayout.LayoutParams(
                    RelativeLayout.LayoutParams.WRAP_CONTENT,
                    RelativeLayout.LayoutParams.WRAP_CONTENT);
            layoutParams.addRule(RelativeLayout.ALIGN_PARENT_LEFT);
            layoutParams.addRule(RelativeLayout.ALIGN_PARENT_TOP);
            layoutParams.width = width;
            layoutParams.height = height;
            layoutParams.leftMargin = x;
            layoutParams.topMargin = y;
            return layoutParams;
        }

        @Override
        protected boolean drawChild(Canvas canvas, View child, long drawingTime) {

            RelativeLayout.LayoutParams layoutParams = (RelativeLayout.LayoutParams) this.getLayoutParams();
            if (cache == null || layoutParams.width != cache.getWidth()
                    || layoutParams.height != cache.getHeight()) {
                this.cache = Bitmap.createBitmap(
                        layoutParams.width, layoutParams.height, Bitmap.Config.ARGB_8888);
            }

            boolean result;
            synchronized(cache){
                /**
                 * the EDT might draw the cache bitmap from within the flushGraphics()
                 * methods. synchronizing here to avoid half painted bitmaps or whatever
                 * might happen in the background if the EDT is reading and we are drawing.
                 */
                result = super.drawChild(new Canvas(this.cache), child, drawingTime);
            }

            /**
             * now that this native component has been painted we certainly need a repaint.
             * notify the EDT.
             */
            Form current = Display.getInstance().getCurrent();
            if(current != null){
               current.repaint(layoutParams.leftMargin, layoutParams.topMargin,
                       layoutParams.width, layoutParams.height);
            }

            return result;
        }

    }
}
