/*
 * Copyright (c) 2008, 2010, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores
 * CA 94065 USA or visit www.oracle.com if you need additional information or
 * have any questions.
 */
package com.sun.lwuit.impl.android;

import com.sun.lwuit.Display;

import android.app.Activity;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Rect;
import android.util.Log;
import android.view.KeyCharacterMap;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.View;
import android.view.SurfaceView;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.view.inputmethod.EditorInfo;
import android.view.inputmethod.InputConnection;
import android.widget.TextView.OnEditorActionListener;
import com.sun.lwuit.Component;
import com.sun.lwuit.TextArea;
import com.sun.lwuit.TextField;

public class AndroidView extends SurfaceView implements SurfaceHolder.Callback{

    private int width = 1;
    private int height = 1;
    private Bitmap bitmap;
    private AndroidGraphics buffy = null;
    private Canvas canvas;
    private AndroidImplementation implementation = null;
    private final Rect bounds = new Rect();
    private boolean fireKeyDown = false;
    private SurfaceHolder surfaceHolder = null;
    private volatile boolean created = false;
    private int deadkey = 0;


    public AndroidView(Activity activity, AndroidImplementation implementation) {
        super(activity);
        this.implementation = implementation;
        setLayoutParams(new ViewGroup.LayoutParams(
                ViewGroup.LayoutParams.FILL_PARENT,
                ViewGroup.LayoutParams.FILL_PARENT));
        setFocusable(true);
        setFocusableInTouchMode(true);
        setEnabled(true);
        setClickable(true);
        setLongClickable(false);

        /**
         * tell the system that we do our own caching and it does not
         * need to use an extra offscreen bitmap.
         */
        setWillNotCacheDrawing(false);

        this.buffy = new AndroidGraphics(implementation, null);

        /**
         * From the docs:
         * "Change whether this view is one of the set of scrollable containers in its window.
         * This will be used to determine whether the window can resize or must pan when a soft
         * input area is open -- scrollable containers allow the window to use resize mode since the
         * container will appropriately shrink. "
         */
        setScrollContainer(true);

        surfaceHolder = getHolder();
        surfaceHolder.addCallback(this);
        android.view.Display androidDisplay = ((WindowManager) activity.getSystemService(Context.WINDOW_SERVICE)).getDefaultDisplay();
        width = androidDisplay.getWidth();
        height = androidDisplay.getHeight();
        initBitmaps(width, height);
        Log.d("LWUIT", "view created");
    }

    //@Override
    public boolean isOpaque() {
        return true;
    }

    @Override
    public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
        if (this.implementation.editInProgress()) {
            /**
             * while edit is in progress a virtual keyboard might
             * resize everything.  this is problematic because the
             * editing (as it is implemented now) blocks the EDT.
             * so we just drop these resize events for now.  once
             * editing is complete we might apply the last state.
             */
            this.implementation.setLastSizeChangedWH(w, h);
            return;
        }

        this.handleSizeChange(w, h);
    }

    @Override
    public void surfaceCreated(SurfaceHolder holder) {
        created = true;
        this.visibilityChangedTo(true);
    }

    @Override
    public void surfaceDestroyed(SurfaceHolder holder) {
        created = false;
        this.visibilityChangedTo(false);
    }

    private void initBitmaps(int w, int h) {
        this.bitmap = Bitmap.createBitmap(w, h, Bitmap.Config.RGB_565);
        this.canvas = new Canvas(this.bitmap);
        this.buffy.setCanvas(this.canvas);
    }

    private void visibilityChangedTo(boolean visible) {
        boolean changed = visible;
        Log.d("LWUIT", " visibilityChangedTo " + visible);
        if (this.implementation.getCurrentForm() != null && changed) {
            if (visible) {
                this.implementation.showNotifyPublic();
                /**
                 * request a full repaint as our surfaceview is most likely
                 * black if this app comes back from the background.
                 */
                this.implementation.getCurrentForm().repaint();
                //android.os.Debug.startMethodTracing("calc");
                Log.d("LWUIT", " repaint ");
            } else {
                this.implementation.hideNotifyPublic();
                //android.os.Debug.stopMethodTracing();
            }
        }
        flushGraphics();
    }

    /*
    @Override
    protected void onWindowVisibilityChanged(int visibility) {
        // method used for View implementation. is it still
        // required with a SurfaceView?
        super.onWindowVisibilityChanged(visibility);
        Log.d("LWUIT", "visibility: " + visibility + " " + View.VISIBLE);
        this.visibilityChangedTo(visibility == View.VISIBLE);
    }
     * 
     */

    @Override
    protected void onSizeChanged(int w, int h, int oldw, int oldh) {
        super.onSizeChanged(w, h, oldw, oldh);

        if (this.implementation.editInProgress()) {
            /**
             * while edit is in progress a virtual keyboard might
             * resize everything.  this is problematic because the
             * editing (as it is implemented now) blocks the EDT.
             * so we just drop these resize events for now.  once
             * editing is complete we might apply the last state.
             */
            this.implementation.setLastSizeChangedWH(w, h);
            return;
        }

        this.handleSizeChange(w, h);
    }

    public void handleSizeChange(int w, int h) {
        if (this.width < w || this.height < h) {
            this.initBitmaps(w, h);
        }
        this.width = w;
        this.height = h;
        Log.d("LWUIT", "sizechanged: " + width + " " + height);
        if (this.implementation.getCurrentForm() == null) {
            /**
             * make sure a form has been set before we can send
             * events to the EDT.  if we send events before the
             * form has been set we might deadlock!
             */
            return;
        }
        Display.getInstance().sizeChanged(w, h);
    }

    @Override
    protected void onDraw(Canvas canvas) {
        boolean empty = canvas.getClipBounds(bounds);
        if (empty) {
            // ??
            canvas.drawBitmap(bitmap, 0, 0, null);
        } else {
            bounds.intersect(0, 0, width, height);
            canvas.drawBitmap(bitmap, bounds, bounds, null);
        }
        if (!implementation.nativePeers.isEmpty()) {
            synchronized (implementation.nativePeers) {
                final int size = implementation.nativePeers.size();
                for (int i = 0; i < size; i++) {
                    AndroidImplementation.AndroidPeer next = (AndroidImplementation.AndroidPeer) implementation.nativePeers.get(i);
                    if (next.layoutWrapper != null) {
                        Bitmap cache = next.layoutWrapper.cache;
                        if (cache != null) {
                            synchronized(cache){
                                canvas.drawBitmap(cache, next.getAbsoluteX(), next.getAbsoluteY(), null);
                            }
                        }
                    }
                }
            }
        }
    }

    public void flushGraphics(Rect rect) {
        if (!created) {
            return;
        }
        Canvas c = null;
        try {
            c = this.surfaceHolder.lockCanvas(rect);
            if (c != null) {
                this.onDraw(c);
            }
        } catch (Throwable e) {
            Log.e("LWUIT", "paint problem.", e);
        } finally {
            if (c != null) {
                this.surfaceHolder.unlockCanvasAndPost(c);
            }
        }
    }

    public void flushGraphics() {
        if (!created) {
            return;
        }
        Canvas c = null;
        try {
            c = this.surfaceHolder.lockCanvas();
            if (c != null) {
                this.onDraw(c);
            }
        } catch (Throwable e) {
            Log.e("LWUIT", "paint problem.", e);
        } finally {
            if (c != null) {
                this.surfaceHolder.unlockCanvasAndPost(c);
            }
        }
    }

    @Override
    public boolean onKeyMultiple(int keyCode, int repeatCount, KeyEvent event) {
        return true;
    }

    @Override
    public boolean onKeyDown(int keyCode, KeyEvent event) {
        this.onKeyUpDown(true, keyCode, event);
        return true;
    }

    @Override
    public boolean onKeyUp(int keyCode, KeyEvent event) {
        this.onKeyUpDown(false, keyCode, event);
        return true;
    }

    /**
     * some info from the MIDP docs about keycodes:
     *
     *  "Applications receive keystroke events in which the individual keys are named within a space of key codes.
     * Every key for which events are reported to MIDP applications is assigned a key code. The key code values are
     * unique for each hardware key unless two keys are obvious synonyms for each other. MIDP defines the following
     * key codes: KEY_NUM0, KEY_NUM1, KEY_NUM2, KEY_NUM3, KEY_NUM4, KEY_NUM5, KEY_NUM6, KEY_NUM7, KEY_NUM8, KEY_NUM9,
     * KEY_STAR, and KEY_POUND. (These key codes correspond to keys on a ITU-T standard telephone keypad.) Other
     * keys may be present on the keyboard, and they will generally have key codes distinct from those list above.
     * In order to guarantee portability, applications should use only the standard key codes.
     *
     * The standard key codes' values are equal to the Unicode encoding for the character that represents the key.
     * If the device includes any other keys that have an obvious correspondence to a Unicode character, their key
     * code values should equal the Unicode encoding for that character. For keys that have no corresponding Unicode
     * character, the implementation must use negative values. Zero is defined to be an invalid key code."
     *
     * Because the MIDP implementation is our reference and that implementation does not interpret the given keycodes
     * we behave alike and pass on the unicode values.
     */
    static int internalKeyCodeTranslate(int keyCode) {
        /**
         * make sure these important keys have a negative value when passed
         * to LWUIT or they might be interpreted as characters.
         */
        switch (keyCode) {
            case KeyEvent.KEYCODE_DPAD_DOWN:
                return AndroidImplementation.DROID_IMPL_KEY_DOWN;
            case KeyEvent.KEYCODE_DPAD_UP:
                return AndroidImplementation.DROID_IMPL_KEY_UP;
            case KeyEvent.KEYCODE_DPAD_LEFT:
                return AndroidImplementation.DROID_IMPL_KEY_LEFT;
            case KeyEvent.KEYCODE_DPAD_RIGHT:
                return AndroidImplementation.DROID_IMPL_KEY_RIGHT;
            case KeyEvent.KEYCODE_DPAD_CENTER:
                return AndroidImplementation.DROID_IMPL_KEY_FIRE;
            case KeyEvent.KEYCODE_MENU:
                return AndroidImplementation.DROID_IMPL_KEY_MENU;
            case KeyEvent.KEYCODE_CLEAR:
                return AndroidImplementation.DROID_IMPL_KEY_CLEAR;
            case KeyEvent.KEYCODE_DEL:
                return AndroidImplementation.DROID_IMPL_KEY_BACKSPACE;
            case KeyEvent.KEYCODE_BACK:
                return AndroidImplementation.DROID_IMPL_KEY_BACK;
            case KeyEvent.KEYCODE_SYM:
                return AndroidImplementation.DROID_IMPL_KEY_SYMBOL;
            default:
                return keyCode;
        }
    }

    private final void onKeyUpDown(boolean down, int keyCode, KeyEvent event) {
        if (event.getRepeatCount() > 0) {
            // skip repeats
            return;
        }

        if (this.implementation.getCurrentForm() == null) {
            /**
             * make sure a form has been set before we can send
             * events to the EDT.  if we send events before the
             * form has been set we might deadlock!
             */
            return;
        }

        keyCode = internalKeyCodeTranslate(keyCode);

        if (keyCode == AndroidImplementation.DROID_IMPL_KEY_FIRE) {
            this.fireKeyDown = down;
        } else if (keyCode == AndroidImplementation.DROID_IMPL_KEY_DOWN
                || keyCode == AndroidImplementation.DROID_IMPL_KEY_UP
                || keyCode == AndroidImplementation.DROID_IMPL_KEY_LEFT
                || keyCode == AndroidImplementation.DROID_IMPL_KEY_RIGHT) {
            if (this.fireKeyDown) {
                /**
                 * we keep track of trackball press/release.  while it is pressed we drop directional
                 * movements.  these movements are most likely not intended.  if the device has no
                 * trackball i see no situation where this additional behavior could hurt.
                 */
                return;
            }

            Log.d("LWUIT", "key: " + keyCode + " down: " + down);

        }

        switch (keyCode) {
            case AndroidImplementation.DROID_IMPL_KEY_BACK:
                if (Display.getInstance().getDefaultVirtualKeyboard().isVirtualKeyboardShowing()) {
                	Display.getInstance().getDefaultVirtualKeyboard().showKeyboard(false);
                    break;
                }
            case AndroidImplementation.DROID_IMPL_KEY_DOWN:
            case AndroidImplementation.DROID_IMPL_KEY_UP:
            case AndroidImplementation.DROID_IMPL_KEY_LEFT:
            case AndroidImplementation.DROID_IMPL_KEY_RIGHT:
            case AndroidImplementation.DROID_IMPL_KEY_FIRE:
            case AndroidImplementation.DROID_IMPL_KEY_MENU:
            case AndroidImplementation.DROID_IMPL_KEY_CLEAR:
            case AndroidImplementation.DROID_IMPL_KEY_BACKSPACE:
            case AndroidImplementation.DROID_IMPL_KEY_SYMBOL:
                // directly pass to display.
                if (down) {
                    Display.getInstance().keyPressed(keyCode);
                } else {
                    Display.getInstance().keyReleased(keyCode);
                }
                break;
            case KeyEvent.KEYCODE_SHIFT_LEFT:
            case KeyEvent.KEYCODE_SHIFT_RIGHT:
            case KeyEvent.KEYCODE_ALT_LEFT:
            case KeyEvent.KEYCODE_ALT_RIGHT:
                // skip
                break;
            default:

                /**
                 * LWUIT's TextField does not seem to work well if two keyup-keydown
                 * sequences of different keys are not strictly sequential.  so we
                 * pass the up event of a character right after the down event.  this is
                 * exactly the behavior of the BlackBerry implementation from this repository
                 * and has worked well for me.  i guess this should be changed as soon as
                 * the TextField changes.
                 */
                if (down) {
                    int nextchar = event.getUnicodeChar();
                    if(nextchar != 0){
                        if((nextchar & KeyCharacterMap.COMBINING_ACCENT) != 0){
                            deadkey = nextchar;
                            break;
                        } else {
                            if(deadkey != 0){
                                nextchar = KeyEvent.getDeadChar(deadkey, keyCode);
                                deadkey = 0;
                            }
                        }
                        Display.getInstance().keyPressed(nextchar);
                        Display.getInstance().keyReleased(nextchar);
                    } else {
                        Log.d("LWUIT", "unrecognized keyCode: " + keyCode);
                    }
                }

                break;
        }
    }

    @Override
    public boolean onTouchEvent(MotionEvent event) {

        if (this.implementation.getCurrentForm() == null) {
            /**
             * make sure a form has been set before we can send
             * events to the EDT.  if we send events before the
             * form has been set we might deadlock!
             */
            return true;
        }

        switch (event.getAction()) {
            case MotionEvent.ACTION_DOWN:
                this.implementation.pointerPressed((int) event.getX(), (int) event.getY());
                break;
            case MotionEvent.ACTION_UP:
                this.implementation.pointerReleased((int) event.getX(), (int) event.getY());
                break;
            case MotionEvent.ACTION_MOVE:
                this.implementation.pointerDragged((int) event.getX(), (int) event.getY());
                break;
        }

        return true;
    }

    public AndroidGraphics getGraphics() {
        return buffy;
    }

    public int getViewHeight() {
        return height;
    }

    public int getViewWidth() {
        return width;
    }


    @Override
    public InputConnection onCreateInputConnection(EditorInfo editorInfo) {

        if(!Display.isInitialized() || Display.getInstance().getCurrent() == null){
            return super.onCreateInputConnection(editorInfo);
        }

        /**
         * do not use the enter key to fire some kind of action!
         */
//        editorInfo.imeOptions |= EditorInfo.IME_ACTION_NONE;

            Component txtCmp = Display.getInstance().getCurrent().getFocused();
            if (txtCmp != null && txtCmp instanceof TextArea) {
                TextArea txt = (TextArea) txtCmp;
                if(txt.isSingleLineTextArea()){
                    editorInfo.imeOptions |= EditorInfo.IME_ACTION_DONE;

                }else{
                    editorInfo.imeOptions |= EditorInfo.IME_ACTION_NONE;
                }
                int inputType = 0;
                switch (txt.getConstraint()) {
                    case TextArea.NUMERIC:
                        inputType = EditorInfo.TYPE_CLASS_NUMBER | EditorInfo.TYPE_NUMBER_FLAG_SIGNED;
                        break;
                    case TextArea.DECIMAL:
                        inputType = EditorInfo.TYPE_CLASS_NUMBER | EditorInfo.TYPE_NUMBER_FLAG_DECIMAL;
                        break;
                    case TextArea.PASSWORD:
                        inputType = EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_PASSWORD;
                        break;
                    case TextArea.PHONENUMBER:
                        inputType = EditorInfo.TYPE_CLASS_PHONE;
                        break;
                    case TextArea.EMAILADDR:
                        inputType = EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_EMAIL_ADDRESS;
                        break;
                    case TextArea.URL:
                        inputType = EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_URI;
                        break;
                    default:
                        inputType = EditorInfo.TYPE_CLASS_TEXT;
                        break;

                }

                editorInfo.inputType = inputType;
                return new LWUITInputConnection(this);
            } else {
                return super.onCreateInputConnection(editorInfo);
            }
    }

    @Override
    public boolean onCheckIsTextEditor() {
        if(!Display.isInitialized() || Display.getInstance().getCurrent() == null){
            return false;
        }
        Component txtCmp = Display.getInstance().getCurrent().getFocused();
        if(txtCmp != null && txtCmp instanceof TextField){
            return true;
        }
        return false;
    }


}
