/*
 * Copyright (c) 2008, 2010, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores
 * CA 94065 USA or visit www.oracle.com if you need additional information or
 * have any questions.
 */
package com.sun.lwuit.impl.android;

import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;

import android.app.Activity;
import android.os.Bundle;
import android.util.Log;
import android.view.Window;

/**
 * It is not obvious how to map the midlet life cycle to the android one.  It seems
 * that we have to consider multiple issues:
 *
 * 1) possibly multiple instances of this activity class exist at the same time.
 *
 * 2) there is no garantee that the onDestroy() method is called but using the onPaused()
 *    method to trigger the midlet destruction is not really an option.
 * 
 * 3) the activity (and in its current implementation the midlet) might be destroyed
 *    but the process might still be alive on the next invocation.  static
 *    variables would survive in this case and cause unforeseen results.
 *
 * I think (1) is addressed in this implementation.  We never really rely on a activity and
 * use it only to initialize views.
 *
 * I see no solution for (2).
 *
 * To avoid problem (3) it seems that a solution would be to use the android:process attribute
 * within the AndroidManifest.xml file and use a value starting with a colon ':' .
 * http://developer.android.com/guide/topics/manifest/activity-element.html#proc
 */
public abstract class LWUITActivity extends Activity {

    private static MIDlet midlet = null;

    /**
     * extend this class and implement this method.  the path
     * returned should point to your midlet class such as:
     * com.padersync.MyMidlet.
     *
     * make your implementation of this activity the app's main
     * activity in the android AndroidManifest.xml file.
     */
    public abstract String getFullApplicationPath();
    /**
     * keep track of current activity instance.
     */
    public static LWUITActivity currentActivity = null;

    @Override
    public void onCreate(Bundle savedInstanceState) {
        Log.d("LWUIT", "onCreate");
        super.onCreate(savedInstanceState);

        if (currentActivity != null) {
            Log.w("LWUIT", "previous activity found!!!");
            this.shootSelf();
            return;
        }

        requestWindowFeature(Window.FEATURE_NO_TITLE);
        currentActivity = this;
        // uncomment for fullscreen.
        //getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,
        //	WindowManager.LayoutParams.FLAG_FULLSCREEN);


        if (midlet == null) {
            try {
                LWUITActivity.currentActivity = LWUITActivity.this;
                midlet = (MIDlet) Class.forName(getFullApplicationPath()).newInstance();
            } catch (Throwable e) {
                Log.e("LWUIT", "FAILED TO CREATE MIDLET " + e.toString());
                this.shootSelf();
            }
        }
    }

    @Override
    protected void onStart() {
        Log.d("LWUIT", "onStart");
        super.onStart();
    }

    @Override
    protected void onDestroy() {
        Log.d("LWUIT", "onDestroy");
        super.onDestroy();
        try {
            midlet.doDestroyApp(true);
        } catch (MIDletStateChangeException e) {
            Log.e("LWUIT", "Problem destroying midlet.",e);
        }
        this.shootSelf();
    }

    @Override
    protected void onPause() {
        Log.d("LWUIT", "onPause");
        super.onPause();

        try {
            midlet.doPauseApp();
        } catch (MIDletStateChangeException e) {
            Log.e("LWUIT", "Problem pausing midlet.",e);
        }

    }

    @Override
    protected void onResume() {
        Log.d("LWUIT", "onResume");
        super.onResume();
        /**
         * spawn a new thread because we need to wait for the
         * views to become visible during initialization and we cannot block this
         * thread.
         */
        new Thread(new Runnable() {

            public void run() {
                try {
                    midlet.doStartApp();
                } catch (MIDletStateChangeException e) {
                    Log.e("LWUIT", "Midlet start failed.", e);
                }
            }
        }).start();

    }

    @Override
    protected void onRestart() {
        Log.d("LWUIT", "onRestart");
        super.onRestart();
    }

    @Override
    protected void onStop() {
        Log.d("LWUIT", "onStop");
        super.onStop();
    }

    /**
     * idea is to make sure the process really ends after this activity
     * is gone.
     */
    private void shootSelf() {
        new Thread(new Runnable() {

            public void run() {
                try {
                    Thread.sleep(1000);
                    android.os.Process.killProcess(android.os.Process.myPid());
                } catch (Throwable e) {
                    Log.e("LWUIT", "problem killing self.", e);
                }
            }
        }).start();
    }
}















